import pygame
import random
import gardener
import euclid

STEP_SIZE = 10

DWARF = "Dwarf"
FLOWER = "Flower"

DWARF_TAG = (0, 0, 0, 255)
FLOWER_TAG = (255, 255, 255, 255)
OBSTACLES_TAG = (0, 0, 0, 255)

class RectSet:
    '''A set of rects without duplicates'''
    def __init__( self ):
        self.rects = []
        return

    def add( self, rect ):
        '''Add the given rect only if it is not already in the set'''
        for item in self.rects:
            if rect == item:
                return
        self.rects.append( rect )
        return

    def point_in_rect( self, point, rect ):
        '''True if the given point is inside the given rect, otherwise False'''
        if point[0]<rect[0] or point[0]>rect[0]+rect[2]:
            return False
        if point[1]<rect[1] or point[1]>rect[1]+rect[3]:
            return False
        return True

    def get_rect( self, pos ):
        '''Return the rect, the given point lies in or None'''
        for rect in self.rects:
            if self.point_in_rect( pos, rect ):
                return rect
        return None

class Item:
    def __init__( self, name, center, width, height, surface ):
        self.name = name
        self.center = center
        self.width = width
        self.height = height
        self.surface = pygame.transform.scale( surface, (width,height) )
        return
    
    def point_in_item( self, point ):
        '''True if the given point is inside the item, otherwise False'''
        rect = self.rect()
        if point[0]<rect[0] or point[0]>rect[0]+rect[2]:
            return False
        if point[1]<rect[1] or point[1]>rect[1]+rect[3]:
            return False
        return True
    
    def rect( self ):
        return ( self.center[0]-self.width/2, self.center[1]-self.height/2, self.width, self.height )
    
    def draw( self, surface ):
        surface.blit( self.surface, (self.center[0]-self.width/2,self.center[1]-self.height/2) )
        return
    
    def speed( self ):
        return 1.0

class AiManager:
    def __init__( self, dwarf_surface, flower_surface ):
        pygame.init()
        random.seed()
        self.dwarf_surface = dwarf_surface
        self.flower_surface = flower_surface
        self.initialize()
        return

    def add_player( self, player ):
        self.players.append( player )
        return

    def initialize( self ):
        self.players = []
        self.start_positions = {}
        self.dwarf_places = {}
        self.flower_places = {}
        self.flowers = []
        self.dwarfs = []
        self.obstacles = RectSet()
        self.owners = {}
        return

    def draw_dwarfs( self, surface ):
        for dwarf in self.dwarfs:
            dwarf.draw( surface )
        return

    def draw_flowers( self, surface ):
        for flower in self.flowers:
            flower.draw( surface )
        return

    def analyze( self, obstacles_surface, places_surface, items_surface, owners_surface ):
        '''Retrieve all required informations from the surface for the added players'''
        start_tags = []
        dwarf_place_tags = []
        flower_place_tags = []

        for player in self.players:
            start_tags.append( player.playerinfo.start_tag )
            dwarf_place_tags.append( player.playerinfo.dwarf_place_tag )
            flower_place_tags.append( player.playerinfo.flower_place_tag )

        self.start_positions = self.retrieve_start_positions( places_surface, start_tags )
        self.dwarf_places = self.retrieve_rects( places_surface, dwarf_place_tags )
        self.flower_places = self.retrieve_rects( places_surface, flower_place_tags )
        self.owners_surface = owners_surface

        flowers = self.retrieve_all_positions( items_surface, FLOWER_TAG )
        dwarfs = self.retrieve_all_positions( items_surface, DWARF_TAG )
        
        for flower in flowers:
            self.flowers.append( Item(FLOWER,flower,50,50,self.flower_surface) )
        for dwarf in dwarfs:
            self.dwarfs.append( Item(DWARF,dwarf,50,50,self.dwarf_surface) )
            
        # retrieve obstacles
        self.obstacles = self.retrieve_rects( obstacles_surface, [OBSTACLES_TAG] )[OBSTACLES_TAG]
        # add upper boundary
        self.obstacles.add( (-10,-10,obstacles_surface.get_width()+20,10) )
        # add lower boundary
        self.obstacles.add( (-10,obstacles_surface.get_height(),obstacles_surface.get_width()+20,10) )
        # add left boundary
        self.obstacles.add( (-10,-10,10,obstacles_surface.get_height()+20) )
        # add right boundary
        self.obstacles.add( (obstacles_surface.get_width(),-10,10,obstacles_surface.get_height()+20) )

        return
    
    def is_point_passable( self, point ):
        for rect in self.obstacles.rects:
            if self.obstacles.point_in_rect( point, rect ):
                return False
        return True
    
    def assign_new_task( self, player ):
        get_dwarf = False
        get_flower = False
        
        # determine if the garden has room for an additional dwarf
        for place in self.dwarf_places[player.playerinfo.dwarf_place_tag].rects:
            if self.is_free_dwarf_place( place ):
                get_dwarf = True
        
        # determine if the garden has room for an additional flower
        for place in self.flower_places[player.playerinfo.flower_place_tag].rects:
            if self.is_free_flower_place( place ):
                get_flower = True
        
        # cant do both so the decision is random
        if get_dwarf == True and get_flower == True:
            if random.randint(0,1) == 0:
                get_dwarf = False
            else:
                get_flower = False
        
        if get_dwarf:
            player.path.append( self.get_random_opponent_dwarf_position(player.playerinfo) )
        else:
            player.path.append( self.get_random_opponent_flower_position(player.playerinfo) )
        
        return
    
    def is_free_dwarf_place( self, place ):
        place_rect = pygame.Rect( place[0], place[1], place[2], place[3] )
        for dwarf in self.dwarfs:
            dwarf_rect = pygame.Rect( dwarf.center[0]-dwarf.width/2, dwarf.center[1]-dwarf.height/2, dwarf.width, dwarf.height )
            if dwarf_rect.colliderect( place_rect ):
                return False
        return True
    
    def is_free_flower_place( self, place ):
        place_rect = pygame.Rect( place[0], place[1], place[2], place[3] )
        for flower in self.flowers:
            flower_rect = pygame.Rect( flower.center[0]-flower.width/2, flower.center[1]-flower.height/2, flower.width, flower.height )
            if flower_rect.colliderect( place_rect ):
                return False
        return True
    
    def assign_path_to_home_dwarf_place( self, player ):
        tag = player.playerinfo.dwarf_place_tag
        rectSet = self.dwarf_places[tag].rects
        
        for rect in rectSet:
            if self.is_free_dwarf_place( rect ):
                position = ( rect[0]+rect[2]/2, rect[1]+rect[3]/2 )
                player.path = self.get_path_from_to( player.center, position )
                return
                
        # else get the first place
        rect = self.dwarf_places[tag].rects[0]
        position = ( rect[0]+rect[2]/2, rect[1]+rect[3]/2 )
        player.path = self.get_path_from_to( player.center, position )
    
    def assign_path_to_home_flower_place( self, player ):
        tag = player.playerinfo.flower_place_tag
        rectSet = self.flower_places[tag].rects
        
        for rect in rectSet:
            if self.is_free_flower_place( rect ):
                position = ( rect[0]+rect[2]/2, rect[1]+rect[3]/2 )
                player.path = self.get_path_from_to( player.center, position )
                return
        
        # else get the first place
        rect = self.flower_places[tag].rects[0]
        position = ( rect[0]+rect[2]/2, rect[1]+rect[3]/2 )
        player.path = self.get_path_from_to( player.center, position )
    
    def get_path_around_rect( self, p1, p2, rect ):
        left = rect[0]
        top = rect[1]
        right = left + rect[2]
        bottom = top + rect[3]
        
        middle_x = left + (right-left)/2
        middle_y = top + (bottom-top)/2
        
        # does the line from p1 to p2 intersect the rect?
        line = euclid.Line2( euclid.Point2(p1[0],p1[1]), euclid.Point2(p2[0],p2[1]) )
            
        # create lines for the current obstacle
        upper_line = euclid.Line2( euclid.Point2(left,top), euclid.Point2(right,top) )
        lower_line = euclid.Line2( euclid.Point2(left,bottom), euclid.Point2(right,bottom) )
        left_line = euclid.Line2( euclid.Point2(left,top), euclid.Point2(left,bottom) )
        right_line = euclid.Line2( euclid.Point2(right,top), euclid.Point2(right,bottom) )
            
        upper_intersection = line.intersect( upper_line )
        lower_intersection = line.intersect( lower_line )
        right_intersection = line.intersect( right_line )
        left_intersection = line.intersect( left_line )
            
        if upper_intersection is None and lower_intersection is None and right_intersection is None and left_intersection is None:
            return []
        
        intersections = []
        if upper_intersection is not None and upper_intersection.x>=left and upper_intersection.x<=right:
            intersections.append( (upper_intersection.x,upper_intersection.y) )
        if lower_intersection is not None and lower_intersection.x>=left and lower_intersection.x<=right:
            intersections.append( (lower_intersection.x,lower_intersection.y) )
        if left_intersection is not None and left_intersection.y>=top and left_intersection.y<=bottom:
            intersections.append( (left_intersection.x,left_intersection.y) )
        if right_intersection is not None and right_intersection.y>=top and right_intersection.y<=bottom:
            intersections.append( (right_intersection.x,right_intersection.y) )
        
        if intersections.__len__() != 2:
            return []
        
        lineToRect1 = euclid.LineSegment2( euclid.Point2(p1[0],p1[1]), euclid.Point2(intersections[0][0],intersections[0][1]) )
        lineToRect2 = euclid.LineSegment2( euclid.Point2(p1[0],p1[1]), euclid.Point2(intersections[1][0],intersections[1][1]) )
        
        if lineToRect2.length < lineToRect1.length:
            intersections = [ intersections[1], intersections[0] ]
        
        
        
        
            
            
        # if yes then determine the shortest path around the rect
        return []
        
    def get_path_from_to( self, p1, p2 ):
        return [p2]
    
    def take_item( self, pos ):
        for dwarf in self.dwarfs:
            if dwarf.point_in_item( pos ):
                return dwarf
        for flower in self.flowers:
            if flower.point_in_item( pos ):
                return flower
        return None
    
    def get_off_my_lawn( self, player ):
        success = False
        # does the player stand on his own ground
        current_owner_tag = self.owners_surface.get_at(player.center)
        if  current_owner_tag != player.playerinfo.start_tag:
            return success
        
        for other_player in self.players:
            if other_player.id == player.id:
                continue
            if self.owners_surface.get_at(other_player.center) == player.playerinfo.start_tag:
                other_player.item = None
                other_player.center = self.get_start_position( other_player.playerinfo )
                other_player.path = []
                success = True
        return success

    def get_start_position( self, playerinfo ):
        if self.start_positions.has_key(playerinfo.start_tag):
            return self.start_positions[playerinfo.start_tag]
        else:
            return (0,0)

    def get_random_opponent_dwarf_position( self, playerinfo ):
        tag = playerinfo.dwarf_place_tag
        random.shuffle( self.dwarfs )
        for dwarf in self.dwarfs:
            # is the current dwarf on my plate
            rectSet = self.dwarf_places[tag]
            for place in rectSet.rects:
                r1 = pygame.Rect( dwarf.rect() )
                r2 = pygame.Rect( place )
                if r1.colliderect(r2) == False:
                    return dwarf.center
        return (0,0)
    
    def get_random_opponent_flower_position( self, playerinfo ):
        tag = playerinfo.flower_place_tag
        random.shuffle( self.flowers )
        for flower in self.flowers:
            # is the current flower in my garden
            rectSet = self.flower_places[tag]
            for place in rectSet.rects:
                if rectSet.point_in_rect( flower.center, place ) == False:
                    return flower.center
        return (0,0)

    def retrieve_start_positions( self, surface, tags ):
        '''Return a dictionary with a start position for each tag'''
        positions = {}

        for tag in tags:
            rects = self.find_surface_rects(surface,tag)
            if rects.rects.__len__()>0:
                center_x = rects.rects[0][0] + rects.rects[0][2]/2
                center_y = rects.rects[0][1] + rects.rects[0][3]/2
                positions[tag] = ( center_x, center_y )
        return positions

    def retrieve_all_positions( self, surface, tag ):
        positions = []
        rects =  self.find_surface_rects( surface, tag )
        for rect in rects.rects:
            center_x = rect[0] + rect[2]/2
            center_y = rect[1] + rect[3]/2
            positions.append( (center_x,center_y) )
        return positions

    def retrieve_rects( self, surface, tags ):
        '''Return a dictionary with a RectSet for each tag'''
        rectSets = {}
        for tag in tags:
            rectSets[tag] = self.find_surface_rects( surface, tag )
        return rectSets

    def find_surface_rects( self, surface, color ):
        '''Return a RectSet with all surface rects of the given color'''
        rects = RectSet()
        max_x = surface.get_width()
        max_y = surface.get_height()
        for x in range(0,max_x,10):
            for y in range(0,max_y,10):
                if surface.get_at((x,y))==color and rects.get_rect((x,y))==None:
                    rects.add( self.get_rect_at(surface,(x,y)) )
        return rects

    def get_rect_at( self, surface, pos ):
        '''Return a rect (left,top,width,height) that the given point lies in'''
        color = surface.get_at(pos)
        left = pos[0]
        while left>0 and surface.get_at((left,pos[1]))==color:
            left = left - 1
        right = pos[0]
        while right<surface.get_width()-1 and surface.get_at((right,pos[1]))==color:
            right = right + 1
        top = pos[1]
        while top>0 and surface.get_at((pos[0],top))==color:
            top = top - 1
        bottom = pos[1]
        while bottom<surface.get_height()-1 and surface.get_at((pos[0],bottom))==color:
            bottom = bottom + 1
        return (left,top,right-left,bottom-top)

    def get_path( self, from_p, to_p ):
        path = []
        path.append( to_p )
        return path

if __name__ == '__main__':
    pygame.init()
    pygame.display.set_mode( (100,100) )
    surface = pygame.image.load("level01_metadata.png").convert_alpha()

    player_1 = PlayerInfo( PLAYER_1_START_TAG, PLAYER_1_DWARF_PLACE_TAG, PLAYER_1_FLOWER_PLACE_TAG )

    ai = AiManager()
    ai.addPlayer( player_1 )
    ai.analyze( surface )

    pygame.quit()